/*
 * CDPBaseCUDA.cpp
 *
 *  Created on: Jun 29, 2009
 *      Author: msuchard
 */

#include "stdafx.h"
#include <cutil_inline.h>
#include <cuda_runtime_api.h>
#include <cuda.h>

#if defined(MULTI_GPU)
	#include <boost/bind.hpp>
#endif

#include "CDPBaseCUDA.h"
#include "CDPBaseCUDA_kernel.h"

#if defined(MULTI_GPU)
	using namespace boost;
#endif
	
static CUT_THREADPROC WKSamplerThread(TGPUplan *plan){
	
	bool multigpu = plan->Multithread && plan->NumDevices> 1;
	if (multigpu) {
		cutilSafeCall( cudaSetDevice(plan->device));
	}

	int N = plan->N;
	int nchunksize = N;
	if (nchunksize>NCHUNKSIZE) { nchunksize = NCHUNKSIZE;}
	if (plan->Multithread) {
		plan->dX = allocateGPURealMemory(DATA_PADDED_DIM * N);
		plan->dMeanAndSigma = allocateGPURealMemory(PACK_DIM * plan->kT * plan->kJ);
		plan->dRandomNumber = allocateGPURealMemory(N);			
		cudaMemcpy(plan->dX,plan->h_X, N * DATA_PADDED_DIM * SIZE_REAL,cudaMemcpyHostToDevice);
		cudaMemcpy(plan->dMeanAndSigma,plan->h_MeanAndSigma ,plan->kT * PACK_DIM * SIZE_REAL, cudaMemcpyHostToDevice);	
		cudaMemcpy(plan->dRandomNumber,plan->h_Rand, N * SIZE_REAL, cudaMemcpyHostToDevice);
		plan->dComponent = allocateGPUIntMemory(N); //only  for K
		plan->dDensity = allocateGPURealMemory(nchunksize * plan->kT * plan->kJ);
	}
			
	int kCumN = 0;
	for (int iChunk = 0; iChunk * nchunksize < N; iChunk++) {	//added layer to reduce GPU memory usage
		kCumN += nchunksize;
		int kNCurrent = nchunksize;
		if(kCumN > N) {
			kNCurrent = N - iChunk * nchunksize;
		}
		gpuMvNormalPDF(plan->dX+iChunk * nchunksize * DATA_PADDED_DIM,plan->dMeanAndSigma,
			plan->dDensity,DIM, kNCurrent, plan->kT * plan->kJ);
		gpuSampleFromMeasureMedium(plan->dDensity,plan->dRandomNumber+iChunk * nchunksize,
			plan->dComponent+iChunk * nchunksize,kNCurrent,plan->kT,plan->kJ);
	}
	cudaMemcpy(plan->h_Component,plan->dComponent,N * SIZE_INT, cudaMemcpyDeviceToHost);

	if (plan->Multithread) {
		cudaFree(plan->dComponent);
		cudaFree(plan->dDensity);
		cudaFree(plan->dX);
		cudaFree(plan->dMeanAndSigma);
		cudaFree(plan->dRandomNumber);
	}
	CUT_THREADEND;
}

CDPBaseCUDA::CDPBaseCUDA() {
	initializedInstance = 0;
}

CDPBaseCUDA::~CDPBaseCUDA(void) {
	finalize();
}

void CDPBaseCUDA::MakeGPUPlans(int startdevice, int numdevices) {
	//--------------------------------------------- 
	DIM = kD;
	CHD_DIM = DIM * (DIM + 1) / 2;		// Entries in the Cholesky decomposition
	MEAN_CHD_DIM = DIM * (DIM + 3) / 2	+ 2;	// Entries in mean, Cholesky decomposition, logDet and p
	LOGDET_OFFSET = DIM * (DIM + 3) / 2;
	PACK_DIM = 16;
	while (MEAN_CHD_DIM > PACK_DIM) {PACK_DIM += 16;}
	DATA_PADDED_DIM = 8;
	while (DIM > DATA_PADDED_DIM) {DATA_PADDED_DIM += 8;}
	//---------------------------------------------

	//host memory
	hX = (REAL*) calloc(DATA_PADDED_DIM * kN,SIZE_REAL);
	hMeanAndSigma = (REAL*) calloc(PACK_DIM * kT * kJ, SIZE_REAL);
	hComponent = (INT*) calloc(kN * 2, SIZE_INT);
	hRandomNumber = (REAL*) calloc(kN, SIZE_REAL);

	int totaldevices = getGPUDeviceCount();
	int nchunksize = kN / numdevices;
	while (nchunksize * numdevices < kN) {
		nchunksize++;
	}
	int i;
	int kCumN = 0;
	for (i = 0; i < numdevices; i++) {
		kCumN += nchunksize;
		int kNCurrent = nchunksize;
		if(kCumN > kN) {
			kNCurrent = kN - i * nchunksize;
		}
		plans[i].N = kNCurrent;
		plans[i].h_X = hX+i * nchunksize * DATA_PADDED_DIM;
		plans[i].h_MeanAndSigma = hMeanAndSigma;
		plans[i].h_Rand = hRandomNumber + i * nchunksize;
		plans[i].h_Component = hComponent+ i * nchunksize;

		//
		plans[i].kT = kT;
		plans[i].kJ = kJ;
		plans[i].device = i+startdevice;
		plans[i].Multithread = numdevices>1;	//number of virtual threads 
		plans[i].NumDevices = totaldevices;		//number of actual threads
	}

	nchunksize = plans[0].N;
	if (nchunksize>NCHUNKSIZE) { nchunksize = NCHUNKSIZE;}
	if (!plans[0].Multithread) { //in this case, we only do allocate device memory once
		plans[0].dMeanAndSigma = allocateGPURealMemory(PACK_DIM * kT * kJ);
		plans[0].dX = allocateGPURealMemory(DATA_PADDED_DIM * plans[0].N);
		plans[0].dRandomNumber = allocateGPURealMemory(plans[0].N);
		plans[0].dComponent = allocateGPUIntMemory(plans[0].N * 2);

		plans[0].dDensity = allocateGPURealMemory(nchunksize * kT * kJ);
	} else {
		#if defined(MULTI_GPU)
			for (i = 0; i < numdevices; i++) {
				workers[i]->call(bind(cudaMalloc, (void**)((void*)&plans[i].dRandomNumber),SIZE_REAL * plans[i].N));
				workers[i]->call(bind(cudaMalloc, (void**)((void*)&plans[i].dComponent),SIZE_INT*  plans[i].N * 2));
				workers[i]->call(bind(cudaMalloc, (void**)((void*)&plans[i].dX),SIZE_REAL * DATA_PADDED_DIM * plans[i].N)); 
				workers[i]->call(bind(cudaMalloc, (void**)((void*)&plans[i].dMeanAndSigma), SIZE_REAL * PACK_DIM * kT * kJ));
				workers[i]->call(bind(cudaMalloc, (void**)((void*)&plans[i].dDensity),SIZE_REAL * nchunksize * kT * kJ));
			}				
		#endif
	}
}

int CDPBaseCUDA::initializeInstance(
						const int iFirstDevice,
						const int iJ,
						const int iT,
						const int iN,
						const int iD,
						const int iNumberDevices
						) {

	kJ = iJ;
	kT = iT;
	kN = iN;
	kD = iD;
	
	fprintf(stderr,"Attempting to initialize GPU device(s)...\n");
	int totalNumDevices = getGPUDeviceCount();
	if (totalNumDevices == 0) {
		fprintf(stderr,"No GPU devices found!\n");
		exit(-1);
	}

	kMultipleDevices = (iNumberDevices == -1) || (iNumberDevices >= 1);
	#if defined(MULTI_GPU)
		if (totalNumDevices == 1) {
			kMultipleDevices = 0;
		}
	#endif

	if (!kMultipleDevices) {
		if (totalNumDevices <= iFirstDevice) {
			fprintf(stderr,"Fewer than %d devices found!\n",(iFirstDevice+1));
			exit(-1);
		}
		printGPUInfo(iFirstDevice);
		cudaSetDevice(iFirstDevice);
		fprintf(stderr,"Device enabled!\n");
		kDeviceCount = 1;
	} else {
		if (iNumberDevices == -1) {
			kDeviceCount = totalNumDevices - iFirstDevice;
		} else {
			// in multi gpu case, kDeviceCount is the number of GPU to be used,
			// in non-multi-gpu case, kDeviceCount is the number of thread to be used
			#if defined(MULTI_GPU)
				if (totalNumDevices < iNumberDevices + iFirstDevice) {
					fprintf(stderr,"Fewer than %d devices found\n",(iFirstDevice+iNumberDevices));
					exit(-1);
				}
			#endif
			kDeviceCount = iNumberDevices;
		}
		#if defined(MULTI_GPU)
		fprintf(stderr,"Number of device(s) used: %d \n",kDeviceCount);
			workers = (GPUWorker**) malloc(kDeviceCount * sizeof(GPUWorker*));
			for(int i=0; i<kDeviceCount; i++) {
				workers[i] = new GPUWorker(iFirstDevice +i);
				printGPUInfo(iFirstDevice + i);
			}
		#endif
	}
	
	fprintf(stderr,"Attempting to allocate memory...\n");
	kStartDevice = iFirstDevice;
	MakeGPUPlans(kStartDevice,kDeviceCount);
	
	#ifdef CHECK_GPU
		checkCUDAError("Post-initialization");
	#endif

	fprintf(stderr,"Memory successfully allocated.\n");
	initializedInstance = 1;

	return CUDA_SUCCESS;
}

void CDPBaseCUDA::finalize(void) {
	if (plans[0].Multithread) {
		#if defined(MULTI_GPU)
			for(int i=0; i<kDeviceCount; i++) {
				workers[i]->call(bind(cudaFree,plans[i].dX));
				workers[i]->call(bind(cudaFree,plans[i].dMeanAndSigma));
				workers[i]->call(bind(cudaFree,plans[i].dDensity));
				workers[i]->call(bind(cudaFree,plans[i].dComponent));
				workers[i]->call(bind(cudaFree,plans[i].dRandomNumber));
			}
			delete [] workers;
		#else
			cudaThreadExit();
		#endif
	} else {
		cudaFree(plans[0].dDensity);
		cudaFree(plans[0].dX);
		cudaFree(plans[0].dMeanAndSigma);
		cudaFree(plans[0].dRandomNumber);
		cudaFree(plans[0].dComponent);
	}

	free(hX); 
	free(hMeanAndSigma);
	free(hComponent);
	free(hRandomNumber);

	#ifdef CHECK_GPU
		checkCUDAError("Post-finalization");
	#endif

	initializedInstance = 0;
}

int CDPBaseCUDA::initializeData(double** iX) {
	if(!initializedInstance) {
		fprintf(stderr,"Device is not yet initialized!\n");
		exit(-1);
	}

	fprintf(stderr,"Attempting to upload data [%d x %d]...\n",kN,kD);

	REAL* hDatumPtr = hX;

	for(int i=0; i<kN; i++) {
		for(int j=0; j<kD; j++)
			hDatumPtr[j] = (REAL) iX[i][j];
		hDatumPtr += DATA_PADDED_DIM;
	}
	
	if (!plans[0].Multithread) {
		cudaMemcpy(plans[0].dX,hX,kN * DATA_PADDED_DIM * SIZE_REAL,cudaMemcpyHostToDevice);
	} else {
		#if defined(MULTI_GPU)
			for(int i=0; i<kDeviceCount; i++) {
				// Copy potion of data onto each GPU
				workers[i]->call(bind(cudaMemcpy,plans[i].dX,plans[i].h_X,
					plans[i].N * DATA_PADDED_DIM * SIZE_REAL, cudaMemcpyHostToDevice)); 
			}
		#endif
	}

	#ifdef CHECK_GPU
		checkCUDAError("Post-data upload");
	#endif

	fprintf(stderr,"Data uploaded!\n");

	return CUDA_SUCCESS;
}

int CDPBaseCUDA::uploadRandomNumbers(MTRand& mt) {

	// TODO Generate random numbers on GPU using MT project in SDK
	for(int i=0; i<kN; i++)
		hRandomNumber[i] = (REAL) mt();
	if (!plans[0].Multithread) {
		cudaMemcpy(plans[0].dRandomNumber,hRandomNumber,kN * SIZE_REAL, cudaMemcpyHostToDevice);
	} else {
		#if defined(MULTI_GPU)
			for(int i=0; i<kDeviceCount; i++) {
				workers[i]->call(bind(cudaMemcpy,plans[i].dRandomNumber,plans[i].h_Rand,
									  plans[i].N * SIZE_REAL, cudaMemcpyHostToDevice));
			}
		#endif
	}
	return CUDA_SUCCESS;
}

int CDPBaseCUDA::uploadMeanAndSigma(
		// TODO Update to use q (J)
		RowVector& p,
		vector<RowVector>& mu,
		vector<UpperTriangularMatrix>& Sigma,
		vector<double>& logdet) {

	REAL* hTmpPtr = hMeanAndSigma;
	for(int i=0; i<kT; i++) {
		double* tMu = mu[i].Store();
		double* tSigma = Sigma[i].Store();

		for(int j=0; j<DIM; j++)
			*hTmpPtr++ = (REAL) *tMu++;

		for(int j=0; j<CHD_DIM; j++)
			*hTmpPtr++ = (REAL) *tSigma++;

		*hTmpPtr++ = (REAL) p[i];
		*hTmpPtr++ = (REAL) logdet[i];

		hTmpPtr += ((PACK_DIM) - (MEAN_CHD_DIM));
	}
	if (!plans[0].Multithread) {
		cudaMemcpy(plans[0].dMeanAndSigma,hMeanAndSigma,kT * PACK_DIM * SIZE_REAL, cudaMemcpyHostToDevice);	
	} else {
		#if defined(MULTI_GPU)
			for(int i=0; i<kDeviceCount; i++) {
				workers[i]->call(bind(cudaMemcpy,plans[i].dMeanAndSigma,plans[i].h_MeanAndSigma,
					kT * kJ * PACK_DIM * SIZE_REAL, cudaMemcpyHostToDevice));
			}	
		#endif
	}
	return CUDA_SUCCESS;
}

int CDPBaseCUDA::sampleWK(
				RowVector& iQ,
				vector<RowVector>& iP,
				vector<RowVector>& iMu,
				vector<UpperTriangularMatrix>& iSigma,
				vector<double>& iLogDet,
				MTRand& mt,
				int *oW,
				int *oK) {

	uploadMeanAndSigma(iP[0], iMu,iSigma,iLogDet);
	uploadRandomNumbers(mt);
	
	if (!plans[0].Multithread) {
		WKSamplerThread(&plans[0]); //no new thread
	} else {
		#if defined(MULTI_GPU)
			int plan;
			for (plan = 0; plan < kDeviceCount; plan++) {
				workers[plan]->callAsync(bind(gpuMvNormalPDF,plans[plan].dX,
							plans[plan].dMeanAndSigma,plans[plan].dDensity,DIM, plans[plan].N,kT * kJ));
			}
			for(plan=0; plan<kDeviceCount; plan++) {
				workers[plan]->call(bind(cudaThreadSynchronize));
			}
			for (plan = 0; plan < kDeviceCount; plan++) {
				workers[plan]->callAsync(bind(gpuSampleFromMeasureMedium,plans[plan].dDensity,
					plans[plan].dRandomNumber,plans[plan].dComponent,plans[plan].N,kT,kJ));
			}
			for(plan=0; plan<kDeviceCount; plan++) {
				workers[plan]->call(bind(cudaThreadSynchronize));
			}
			for(plan=0; plan<kDeviceCount; plan++) {
				workers[plan]->call(bind(cudaMemcpy,plans[plan].h_Component,plans[plan].dComponent,plans[plan].N * SIZE_INT, cudaMemcpyDeviceToHost));
			}
		#else
			for (int plan = 0; plan < kDeviceCount; plan++) {
				threadID[plan] = cutStartThread((CUT_THREADROUTINE)WKSamplerThread, (void *)(plans + plan));
			}
			cutWaitForThreads(threadID, kDeviceCount);
		#endif
	}
	
	memcpy(oK, hComponent, kN * SIZE_INT);
	return CUDA_SUCCESS;
}

int CDPBaseCUDA::getGPUDeviceCount() {
	int cDevices;
	CUresult status;
	status = cuInit(0);
	if (CUDA_SUCCESS != status)
		return 0;
	status = cuDeviceGetCount(&cDevices);
	if (CUDA_SUCCESS != status)
		return 0;
	if (cDevices == 0) {
		return 0;
	}
	return cDevices;
}

void CDPBaseCUDA::printGPUInfo(int iDevice) {

	fprintf(stderr,"GPU Device Information:");

		char name[256];
		int totalGlobalMemory = 0;
		int clockSpeed = 0;

		// New CUDA functions in cutil.h do not work in JNI files
		getGPUInfo(iDevice, name, &totalGlobalMemory, &clockSpeed);
		fprintf(stderr,"\nDevice #%d: %s\n",(iDevice+1),name);
		double mem = totalGlobalMemory / 1024.0 / 1024.0;
		double clo = clockSpeed / 1000000.0;
		fprintf(stderr,"\tGlobal Memory (MB) : %3.0f\n",mem);
		fprintf(stderr,"\tClock Speed (Ghz)  : %1.2f\n",clo);
}

void CDPBaseCUDA::getGPUInfo(int iDevice, char *oName, int *oMemory, int *oSpeed) {
	cudaDeviceProp deviceProp;
	memset(&deviceProp, 0, sizeof(deviceProp));
	cudaGetDeviceProperties(&deviceProp, iDevice);
	*oMemory = deviceProp.totalGlobalMem;
	*oSpeed = deviceProp.clockRate;
	strcpy(oName, deviceProp.name);
}

